// LALR    () Proteus   lawnmower-man@mail.ru

#include<stdio.h>
#include<assert.h>
#include<vector>

typedef unsigned char byte;

//!  () LALR 
enum{
	END=1,LINK,
	SET,CHAR,SET_CHAR,ALL,ITEM,LIST,ATOM,SET1,S_ATOM
};

//!  ()
struct Node{
	enum e{Sym,Or,OneMore,ZeroMore,ZeroOne,Accept,Con}op;
	Node *p1,*p2;
	byte sym1,sym2;
	Node(char s1,char s2):op(Sym),sym1(s1),sym2(s2),p1(0),p2(0) {}
	Node(e p,Node *n1=0,Node *n2=0):op(p),sym1(0),sym2(0),p1(n1),p2(n2) {}
	Node *Link(Node *n) {
		if(p2) return new Node(Con,this,n);
		p2=n; return this;
	}
};

//!  LR 
struct StackItem{
	byte index;		//!<   id ,      
	byte sym;		//!<  
	Node *node;		//!<  
};

//!    (    )
const struct GramItem{byte size; byte sym;}gram_list[]={
	{0,0},
	{2,ALL},	//  1  all: list
	{4,ALL},	//  2  all: all | list
	{2,LIST},	//  3  list: item
	{3,LIST},	//  4  list: item list
	{2,ITEM},	//  5  item: atom
	{3,ITEM},	//  6  item: atom +
	{3,ITEM},	//  7  item: atom *
	{3,ITEM},	//  8  item: atom ?
	{2,ATOM},	//  9  atom: char
	{2,ATOM},	// 10  atom: set
	{4,ATOM},	// 11  atom: ( all )
	{4,SET},	// 12  set: [ set1 ]
	{2,SET1},	// 13  set1: set_atom
	{3,SET1},	// 14  set1: set1 set_atom
	{2,S_ATOM},	// 15  set_atom: char
	{4,S_ATOM}	// 16  set_atom: char - char 
};

//!   
const struct State{byte sym; char action;}table[]={
					// state 0 index 0
	{ALL,27},		// goto=8
	{LINK,6},
					// state 1 index 2
	{S_ATOM,-13},	// Ireduce!=13
	{SET1,20},		// goto=10
	{LINK,22},
					// state 2 index 5
	{ALL,26},		// goto=12
	{LIST,-1},		// Ireduce!=1	<= index 6
	{LINK,13},
					// state 5 index 8
	{'?',-8},		// Ireduce!=8
	{'*',-7},		// Ireduce!=7
	{'+',-6},		// Ireduce!=6
	{0,-5},			// reduce=5
					// state 6 index 12
	{LIST,-4},		// Ireduce!=4
	{'[',2},		// goto=1		<=  index 13
	{'(',5},		// goto=2
	{CHAR,-9},		// Ireduce!=9
	{SET,-10},		// Ireduce!=10
	{ATOM,8},		// goto=5
	{ITEM,12},		// goto=6
	{0,-3},			// reduce 3
					// state 8 index 20
//	{END,0},
//	{LINK,27},
					// state 10 index 20
	{S_ATOM,-14},	// Ireduce!=14
	{']',-12},		// Ireduce!=12
	{SET_CHAR,24},	// goto=11		<= index 22
	{0,0},
					// state 11 index 24
	{'-',31},		// goto=20
	{0,-15},		// reduce=15
					// state 12 index 26
	{')',-11},		// Ireduce!=11
	{'|',29},		// goto=17		<= index 27
	{0,0},
					// state 17 index 29
	{LIST,-2},		// Ireduce!=2
	{LINK,13},
					// state 20 index 31
	{CHAR,-16},		// Ireduce!=16
	{0,0}
					// index 33
};


//!  
class Parser{
	StackItem curr_state;			//!<   LALR 
	std::vector<StackItem>stack;
	bool Step(byte);				//!<   
	void Reduce(int);				//!<   
	void Action(StackItem*,int);	//!<   
	const State *FindStatus();		//!<     
public:
	Node *BuildTree(const char*);	//!<      
	Node *BuildGraph(Node*,Node *f=new Node(Node::Accept)); //!<    
};


//    
Node *Parser::BuildGraph(Node *start,Node *final)
{
	Node *t,*f;
	if(!start) return final;
	f=BuildGraph(start->p2,final);
	switch(start->op) {
	//case Node::Accept: break;
	case Node::Con: 
		assert(start->p2);
		start=BuildGraph(start->p1,f);
		break;
	case Node::Sym:  
		start->p2=f;
		break;
	case Node::ZeroOne:
		start->p1=BuildGraph(start->p1,f);
		start=new Node(Node::Or,start->p1,f);
		break;
	case Node::OneMore:
		start=BuildGraph(start->p1,t=new Node(Node::Or,f));
		t->p2=start;
		break;
	case Node::ZeroMore:
		start=new Node(Node::Or,start->p1,f);
		start->p1=BuildGraph(start->p1,start);
		break;
	case Node::Or:
		start->p1=BuildGraph(start->p1,final);
		start->p2=f;
		break;
	}
	return start;
}


Node *Parser::BuildTree(const char *text)
{
	memset(&curr_state,0,sizeof curr_state);
	stack.clear();
	stack.push_back(curr_state);
	while(*text) if(Step(*text++)) return 0;
	Step(END);
	return stack.size()==2?stack.back().node:0;
}


bool Parser::Step(byte sym)
{
	curr_state.sym=sym; sym=0;
	for(;;) {
		const State *st=FindStatus(); //    
		if(!st) return true;
		if(st->sym) { // shift
			curr_state.index=st->action;
			stack.push_back(curr_state);
		}
		if(st->action<0) { // reduce
			if(!st->sym) sym=curr_state.sym;
			Reduce(-st->action);
			continue;
		}
		if(!sym) break;
		curr_state.sym=sym; sym=0;
	}
	return false;
}


inline void Parser::Reduce(int gram_index)
{
	int size=gram_list[gram_index].size;
	stack.push_back(curr_state);
	Action(&stack.back(),gram_index);	//  
	curr_state=stack.back();			//  
	stack.resize(stack.size()-size);	//     
	curr_state.index=stack.back().index;
	curr_state.sym=gram_list[gram_index].sym;
}


inline const State *Parser::FindStatus()
{
	bool def=true; const char *except=0;
	int index=curr_state.index;
	byte find_sym=curr_state.sym;
	for(;;) {
		byte sym=table[index].sym;
		switch(sym) {
		case 0:    goto exit;
		case LINK: index=table[index].action; def=false; continue;
		case CHAR:     except="(|)"; break;
		case SET_CHAR: except="]"; break;
		}
		if(sym==find_sym || except && find_sym>' ' && !strchr(except,find_sym)) return &table[index];
		index++;
	}
exit:
	if(def && table[index].action) return &table[index];
	return 0;
}


void Parser::Action(StackItem *sp,int index)
{
	switch(index) {
	case 2:  sp->node=new Node(Node::Or,sp[-1].node,sp[-3].node); break;
	case 4:  sp->node=sp[-2].node->Link(sp[-1].node); break;
	case 6:  sp->node=new Node(Node::OneMore ,sp[-2].node); break;
	case 7:  sp->node=new Node(Node::ZeroMore,sp[-2].node); break;
	case 8:  sp->node=new Node(Node::ZeroOne ,sp[-2].node); break;
	case 9:
	case 15: sp->node=new Node(sp[-1].sym,sp[-1].sym); break;
	case 11:
	case 12: sp->node=sp[-2].node; break;
	case 14: sp->node=new Node(Node::Or,sp[-1].node,sp[-2].node); break;
	case 16: sp->node=new Node(sp[-3].sym,sp[-1].sym); break;
	};
}


#include<set>

//  ()  
void Dump(FILE *file,std::set<Node*> &set,Node *node,Node *parent=0)
{
	const char *shape=0; 
	if(!node) return;
	if(parent) fprintf(file,"  %d -> %d\n",parent,node); else shape=",fontcolor=green";
	if(set.find(node)!=set.end()) return; set.insert(node);
	char label[10]; label[1]=0;
	switch(node->op) {
	case Node::Sym:      
		if(node->sym1==node->sym2) label[0]=node->sym1; 
		else sprintf(label,"%c - %c",node->sym1,node->sym2); 
		if(!shape) shape=""; break;
	case Node::Con:      strcpy(label,"Con"); break;
	case Node::Accept:   strcpy(label,"Accept"); shape=",fontcolor=green"; break;
	case Node::Or:       strcpy(label,"Or"); break;
	case Node::OneMore:  label[0]='+'; break; 
	case Node::ZeroOne:  label[0]='?'; break;
	case Node::ZeroMore: label[0]='*'; break;
	default: strcpy(label,"<err>"); break;
	}
	if(!shape) shape=",fontcolor=gray";
	fprintf(file,"  %d [label=\"%s\"%s]\n",node,label,shape);
	Dump(file,set,node->p1,node);
	Dump(file,set,node->p2,node);
}


void Dump(const char *filename,Node *node)
{
	std::set<Node*>set;
	FILE *file=fopen(filename,"wt");
	if(!file) return;
	fprintf(file,
		"# www.graphviz.org\n"
		"digraph G{\n"
		"  node [shape=none]\n"
		"  rankdir=LR\n"
		);
	Dump(file,set,node);
	fprintf(file,"}\n");
	fclose(file);
}


int main()
{
	Parser obj; Node *node;
	node=obj.BuildTree("x*y+z?|[rx-z](ab|c)"); 
	Dump("tree.dot",node);
	node=obj.BuildGraph(node);
	Dump("graph.dot",node);
	return 0;
}

